from distutils.command.build_ext import build_ext
from distutils.core import Distribution
from distutils.errors import (CCompilerError, CompileError, DistutilsExecError,
                              DistutilsPlatformError)
from distutils.extension import Extension
from os import chmod, stat
from pathlib import Path
from shutil import copyfile
from subprocess import run


class GoExtension(Extension):
    def __init__(self, name, path):
        super().__init__(name, sources=[])
        self.path = path


extensions = [
    GoExtension(
        #"groove",
        "groove.assets.grooveproxy",
        # Assume we have temporarily copied over the proxy folder into our current path
        "./proxy",
    )
]


class BuildFailed(Exception):
    pass


class GoExtensionBuilder(build_ext):
    def run(self):
        try:
            build_ext.run(self)
        except (DistutilsPlatformError, FileNotFoundError):
            raise BuildFailed("File not found. Could not compile extension.")

    def build_extension(self, ext):
        try:
            if isinstance(ext, GoExtension):
                extension_root = Path(__file__).parent.resolve() / ext.path
                ext_path = self.get_ext_fullpath(ext.name)
                result = run(["go", "build", "-o", str(Path(ext_path).absolute())], cwd=extension_root)
                if result.returncode != 0:
                    raise CompileError("Go build failed")
            else:
                build_ext.build_extension(self, ext)
        except (CCompilerError, DistutilsExecError, DistutilsPlatformError, ValueError):
            raise BuildFailed('Could not compile C extension.')


def build(setup_kwargs):
    distribution = Distribution({"name": "python_ctypes", "ext_modules": extensions})
    distribution.package_dir = "python_ctypes"

    cmd = GoExtensionBuilder(distribution)
    cmd.ensure_finalized()
    cmd.run()

    for output in cmd.get_outputs():
        relative_extension = Path(output).relative_to(cmd.build_lib)
        copyfile(output, relative_extension)
        mode = stat(relative_extension).st_mode
        mode |= (mode & 0o444) >> 2
        chmod(relative_extension, mode)

if __name__ == "__main__":
    build({})



from collections import Counter, defaultdict
from contextlib import contextmanager
from dataclasses import dataclass
from functools import partial
from itertools import groupby
from multiprocessing import Process, Semaphore
from time import sleep
from typing import Any, Optional

import uvicorn
from fastapi import FastAPI
from fastapi.responses import Response


@dataclass
class MockPageDefinition:
    url: str
    status_code: int = 200
    content_type: str = "text/html"

    method: str = "get"
    headers: dict[str, str] | None = None

    # Optional function callable - one should be specified
    content: Optional[str] = None
    handler: Optional[Any] = None

def launch_server(payloads: list[MockPageDefinition], port: int, setup_callback: Optional[Any], allow_repeat_access: bool, semaphore: Semaphore):
    server = FastAPI()
    requests_by_url = defaultdict(int)

    if setup_callback is not None:
        setup_callback(server)
    
    def render_favicon():
        return "fav"

    def render_default(page_definitions: list[MockPageDefinition]):
        url = page_definitions[0].url
        existing_requests = requests_by_url[url]

        if len(page_definitions) > 1 and existing_requests >= len(page_definitions):
            raise ValueError(f"Accessing beyond bound of sequential definitions: `{url}`.")
        if len(page_definitions) == 1 and existing_requests > 0 and not allow_repeat_access:
            raise ValueError(f"You have already accessed this url: `{url}`.")

        if len(page_definitions) > 1:
            definition = page_definitions[existing_requests]
        else:
            definition = page_definitions[0]

        requests_by_url[url] += 1

        return Response(
            content=definition.content, 
            status_code=definition.status_code,
            media_type=definition.content_type,
            headers=definition.headers,
        )

    # Group by URL (maintaining the original order)
    definitions_by_url = {
        (url, method): [definition for _, definition in definitions]
        for (url, method), definitions in groupby(
            sorted(enumerate(payloads), key=lambda x: (x[1].url, x[1].method, x[0])),
            key=lambda x: (x[1].url, x[1].method)
        )
    }

    for (url, method), definitions in definitions_by_url.items():
        mount_fn = getattr(server, method)

        # Only one URL should be provided if a handler is active
        has_handler = any([definition.handler for definition in definitions])
        if has_handler and len(definitions) > 1:
            raise ValueError(f"Only one definition is supported when handler is active: `{url}`")

        mount_fn(url)(definitions[0].handler or partial(render_default, page_definitions=definitions))

    def not_found_fallback(path):
        raise ValueError(f"No page matching query: {path}")

    server.get("/favicon.ico")(render_favicon)
    # TODO: Switch to a 404 handler
    server.get("{path:path}")(not_found_fallback)
    semaphore.release()
    uvicorn.run(server, host="0.0.0.0", port=port)

@contextmanager
def mock_server(
    payloads: list[MockPageDefinition],
    port=6012,
    setup_callback=None,
    allow_repeat_access: bool = True,
):
    """
    Sets up a test server parameterized by MockPageDefinition in the background.

    :param setup_callback: If specified, will be passed the FastAPI server object once createed. This function
        will be called in the separate process.
    :parma allow_repeat_access: Allow the same URL to be accessed multiple times within the
        same mock session. If false will throw an error if already accessed. Only applies to entries
        that have one definition per URL. If multiple definitions are given per URL, will assume
        that you intend to iterate them.

    """
    process_ready_semaphore = Semaphore(0)

    process = Process(target=launch_server, args=(payloads, port, setup_callback, allow_repeat_access, process_ready_semaphore))
    process.start()

    with process_ready_semaphore:
        try:
            print("Launched mock server...")
            # Wait for process to actually spawn
            sleep(0.1)
            yield f"http://localhost:{port}"
        except Exception as e:
            raise e
        finally:
            process.kill()



from base64 import b64decode, b64encode
from contextlib import contextmanager
from enum import Enum
from gzip import compress, decompress
from json import dumps, loads
from subprocess import Popen
from sysconfig import get_config_var
from time import sleep
from urllib.parse import urljoin

from pydantic import BaseModel, validator
from requests import Session

from groove.assets import get_asset_path


class CacheModeEnum(Enum):
    # Ensure enum values are aligned with the cache.go definitions
    OFF = 0
    STANDARD = 1
    AGGRESSIVE = 2


class TapeRequest(BaseModel):
    url: str
    method: str
    headers: dict[str, list[str]]
    body: bytes

    @validator("body")
    def validate_body(cls, value):
        return b64decode(value)


class TapeResponse(BaseModel):
    status: int
    headers: dict[str, list[str]]
    body: bytes

    @validator("body")
    def validate_body(cls, value):
        return b64decode(value)


class TapeRecord(BaseModel):
    request: TapeRequest
    response: TapeResponse

    class Config:
        json_encoders = {
            # Assume that body bytes should always be encoded as base64 strings
            bytes: lambda value: b64encode(value).decode(),
        }


class TapeSession(BaseModel):
    records: list[TapeRecord]

    @classmethod
    def from_server(cls, data: bytes):
        raw_records = loads(decompress(data))
        return cls(records=raw_records)

    def to_server(self) -> bytes:
        return compress(
            dumps([
                # json_encoders doesn't operate on list items, must iterate manually
                # https://github.com/pydantic/pydantic/issues/4085
                loads(record.json())
                for record in self.records
            ]).encode()
        )


class Groove:
    def __init__(
        self,
        command_timeout: int = 5,
        port: int = 6010,
        control_port: int = 6011,
        proxy_server: str | None = None,
        proxy_username: str | None = None,
        proxy_password: str | None = None,
        auth_username: str | None = None,
        auth_password: str | None = None,
    ):
        self.session = Session()

        self.port = port
        self.control_port = control_port
        self.proxy_server = proxy_server
        self.proxy_username = proxy_username
        self.proxy_password = proxy_password
        self.auth_username = auth_username
        self.auth_password = auth_password

        self.base_url_proxy = f"http://localhost:{port}"
        self.base_url_control = f"http://localhost:{control_port}"

        self.timeout = command_timeout

    @contextmanager
    def launch(self):
        parameters = {
            "--port": self.port,
            "--control-port": self.control_port,
            "--proxy-server": self.proxy_server,
            "--proxy-username": self.proxy_username,
            "--proxy-password": self.proxy_password,
            "--auth-username": self.auth_username,
            "--auth-password": self.auth_password,
        }

        # Not specifying parameters should make them null
        parameters = {key: value for key, value in parameters.items() if value is not None}

        process = Popen(
            [
                self.executable_path,
                *[
                    str(item)
                    for key, value in parameters.items()
                    for item in [key, value]
                ]
            ]
        )

        # Wait for launch
        sleep(0.5)

        try:
            yield
        finally:
            process.terminate()

    def tape_start(self):
        response = self.session.post(urljoin(self.base_url_control, "/api/tape/record"), timeout=self.timeout)
        assert response.json()["success"] == True

    def tape_get(self) -> TapeSession:
        tape_response = self.session.post(urljoin(self.base_url_control, "/api/tape/retrieve"), timeout=self.timeout)
        return TapeSession.from_server(tape_response.content)

    def tape_load(self, session: TapeSession):
        tape_response = self.session.post(
            urljoin(self.base_url_control, "/api/tape/load"),
            files={"file": session.to_server()},
            timeout=self.timeout,
        )
        assert tape_response.json()["success"] == True

    def tape_stop(self):
        response = self.session.post(urljoin(self.base_url_control, "/api/tape/stop"), timeout=self.timeout)
        assert response.json()["success"] == True

    def set_cache_mode(self, mode: CacheModeEnum):
        response = self.session.post(
            urljoin(self.base_url_control, "/api/cache/mode"),
            json=dict(
                mode=mode.value,
            )
        )
        assert response.json()["success"] == True

    @property
    def executable_path(self) -> str:
        # Support statically and dynamically build libraries
        if (path := get_asset_path("grooveproxy")).exists():
            return str(path)

        wheel_extension = get_config_var("EXT_SUFFIX")
        if (path := get_asset_path(f"grooveproxy{wheel_extension}")).exists():
            return exit(path)

        raise ValueError("No groove executable file found")



from base64 import b64encode

import pytest
from bs4 import BeautifulSoup
from playwright._impl._api_types import Error as PlaywrightError
from requests import get

from groove.assets import get_asset_path
from groove.proxy import (Groove, TapeRecord, TapeRequest, TapeResponse,
                          TapeSession)

AUTH_USERNAME = "test-username"
AUTH_PASSWORD = "test-password"


@pytest.mark.xfail()
def test_auth_requests():
    proxy = Groove(port=6040, control_port=6041, auth_username=AUTH_USERNAME, auth_password=AUTH_PASSWORD)

    record = TapeRecord(
        request=TapeRequest(
            url="https://freeman.vc:443/",
            method="GET",
            headers={},
            body=b"",
        ),
        response=TapeResponse(
            status=200,
            headers={},
            body=b64encode(b"Test content")
        ),
    )

    with proxy.launch():
        proxy.tape_load(
            TapeSession(
                records=[record]
            )
        )

        response = get(
            "https://freeman.vc",
            proxies={
                "http": f"http://{proxy.auth_username}:{proxy.auth_password}@localhost:{proxy.port}",
                "https": f"http://{proxy.auth_username}:{proxy.auth_password}@localhost:{proxy.port}",
            },
            verify=get_asset_path("ssl/ca.crt"),
        )
        assert response.ok
        assert BeautifulSoup(response.content).strip() == "Test content"


@pytest.mark.xfail()
def test_auth_chromium(browser):
    """
    Ensure the proxy can forward to an end proxy
    """
    proxy = Groove(port=6040, control_port=6041, auth_username=AUTH_USERNAME, auth_password=AUTH_PASSWORD)

    record = TapeRecord(
        request=TapeRequest(
            url="https://freeman.vc:443/",
            method="GET",
            headers={},
            body=b"",
        ),
        response=TapeResponse(
            status=200,
            headers={},
            body=b64encode(b"Test content")
        ),
    )

    with proxy.launch():
        proxy.tape_load(
            TapeSession(
                records=[record]
            )
        )

        # Make sure the end proxy has configured correctly
        context = browser.new_context(
            proxy={
                "server": proxy.base_url_proxy,
                "username": proxy.auth_username,
                "password": proxy.auth_password,
            },
        )
        page = context.new_page()
        page.goto("https://freeman.vc", timeout=5000)
        assert BeautifulSoup(page.content()).text.strip() == "Test content"



from uuid import uuid4

from bs4 import BeautifulSoup

from groove.proxy import CacheModeEnum
from groove.tests.mock_server import MockPageDefinition, mock_server


def test_cache_off(proxy, browser):
    """
    Ensure the cache is off will route all requests
    """
    proxy.set_cache_mode(CacheModeEnum.OFF)

    # Leverage random identifiers for each test to ensure there isn't
    # data leakage from one unit test to the other
    request1_content = str(uuid4())
    request2_content = str(uuid4())

    with mock_server([
        MockPageDefinition(
            "/test",
            content=f"<html><body>{request1_content}</body></html>"
        ),
        MockPageDefinition(
            "/test",
            content=f"<html><body>{request2_content}</body></html>"
        ),
    ]) as mock_url:
        context = browser.new_context(
            proxy={
                "server": proxy.base_url_proxy,
            }
        )
        page = context.new_page()
        page.goto(f"{mock_url}/test")
        assert BeautifulSoup(page.content()).text.strip() == request1_content

        page.goto(f"{mock_url}/test")
        assert BeautifulSoup(page.content()).text.strip() == request2_content


def test_cache_standard(proxy, browser):
    """
    Ensure the cache respects server headers
    """
    proxy.set_cache_mode(CacheModeEnum.STANDARD)

    request1_content = str(uuid4())
    request2_content = str(uuid4())

    with mock_server([
        MockPageDefinition(
            "/test",
            content=f"<html><body>{request1_content}</body></html>",
            headers={
                "Cache-Control": "max-age=604800",
            }
        ),
        MockPageDefinition(
            "/test",
            content=f"<html><body>{request2_content}</body></html>",
            headers={
                "Cache-Control": "max-age=604800",
            }
        ),
    ]) as mock_url:
        context = browser.new_context(
            proxy={
                "server": proxy.base_url_proxy,
            }
        )
        page = context.new_page()
        page.goto(f"{mock_url}/test")
        assert BeautifulSoup(page.content()).text.strip() == request1_content

        # We should never hit the second definition because of the first requests' headers
        page.goto(f"{mock_url}/test")
        assert BeautifulSoup(page.content()).text.strip() == request1_content


def test_cache_aggressive(proxy, browser):
    """
    Ensure the aggressive cache will cache all requests
    """
    # Clear previous cache records, if they exist
    proxy.set_cache_mode(CacheModeEnum.OFF)
    proxy.set_cache_mode(CacheModeEnum.AGGRESSIVE)

    request1_content = str(uuid4())
    request2_content = str(uuid4())

    with mock_server([
        MockPageDefinition(
            "/test",
            content=f"<html><body>{request1_content}</body></html>"
        ),
        MockPageDefinition(
            "/test",
            content=f"<html><body>{request2_content}</body></html>"
        ),
    ]) as mock_url:
        context = browser.new_context(
            proxy={
                "server": proxy.base_url_proxy,
            }
        )
        page = context.new_page()
        page.goto(f"{mock_url}/test")
        assert BeautifulSoup(page.content()).text.strip() == request1_content

        page.goto(f"{mock_url}/test")
        assert BeautifulSoup(page.content()).text.strip() == request1_content



import pytest
from playwright.sync_api import sync_playwright

from groove.proxy import Groove


# We want this to recreate by default on every unit test to clear the state
@pytest.fixture(scope="function")
def proxy():
    proxy = Groove()
    with proxy.launch():
        yield proxy

@pytest.fixture(scope="function")
def browser():
    with sync_playwright() as p:
        yield p.chromium.launch(
            headless=True,
        )



from base64 import b64encode
from uuid import uuid4

from bs4 import BeautifulSoup

from groove.proxy import TapeRecord, TapeRequest, TapeResponse, TapeSession


def test_tape(proxy, browser):
    """
    Ensure the basic tape functions work correctly
    """
    proxy.tape_start()

    context = browser.new_context(
        proxy={
            "server": proxy.base_url_proxy,
        }
    )
    page = context.new_page()
    page.goto("https://freeman.vc")

    modified_records = 0
    session = proxy.tape_get()
    assert len(session.records) > 0

    for record in session.records:
        if record.request.url == "https://freeman.vc:443/":
            record.response.body = "Mocked content".encode()
            modified_records += 1
    assert modified_records == 1

    proxy.tape_load(session)

    context = browser.new_context(
        proxy={
            "server": proxy.base_url_proxy,
        }
    )
    page = context.new_page()
    page.goto("https://freeman.vc")

    assert BeautifulSoup(page.content()).text.strip() == "Mocked content"


def test_multiple_requests(proxy, browser):
    """
    Ensure mocked requests resolve in the same order
    """
    response_1 = str(uuid4())
    response_2 = str(uuid4())

    records = [
        TapeRecord(
            request=TapeRequest(
                url="https://freeman.vc:443/",
                method="GET",
                headers={},
                body=b"",
            ),
            response=TapeResponse(
                status=200,
                headers={},
                body=b64encode(response.encode())
            ),
        )
        for response in [response_1, response_2]
    ]

    proxy.tape_load(
        TapeSession(
            records=records
        )
    )

    context = browser.new_context(
        proxy={
            "server": proxy.base_url_proxy,
        }
    )
    page = context.new_page()

    page.goto("https://freeman.vc")
    assert BeautifulSoup(page.content()).text.strip() == response_1

    page.goto("https://freeman.vc")
    assert BeautifulSoup(page.content()).text.strip() == response_2



from base64 import b64encode

import pytest
from bs4 import BeautifulSoup
from playwright.sync_api import sync_playwright

from groove.proxy import (Groove, TapeRecord, TapeRequest, TapeResponse,
                          TapeSession)

AUTH_USERNAME = "test-username"
AUTH_PASSWORD = "test-password"


@pytest.mark.parametrize(
    "end_proxy,middle_proxy",
    [
        (
            # Unauthenticated end proxy
            Groove(port=6040, control_port=6041),
            Groove(port=6010, control_port=6011),
        ),
        #(
        #    # Authenticated end proxy
        #    # Currently failing because of Chromium not sending Auth headers on every request
        #    Groove(port=6040, control_port=6041, auth_username=AUTH_USERNAME, auth_password=AUTH_PASSWORD),
        #    Groove(port=6010, control_port=6011, proxy_username=AUTH_USERNAME, proxy_password=AUTH_PASSWORD),
        #)
    ]
)
def test_end_proxy(end_proxy, middle_proxy, browser):
    """
    Ensure the proxy can forward to an end proxy
    """
    middle_proxy.proxy_server = end_proxy.base_url_proxy

    record = TapeRecord(
        request=TapeRequest(
            url="https://freeman.vc:443/",
            method="GET",
            headers={},
            body=b"",
        ),
        response=TapeResponse(
            status=200,
            headers={},
            body=b64encode(b"Test content")
        ),
    )

    with middle_proxy.launch():
        with end_proxy.launch():
            end_proxy.tape_load(
                TapeSession(
                    records=[
                        # Double requests for the two proxy requests
                        record,
                        record,
                    ]
                )
            )

            proxy_payload = {
                "server": end_proxy.base_url_proxy,
                **(
                    {
                        "username": end_proxy.auth_username,
                        "password": end_proxy.auth_password,
                    }
                    if end_proxy.auth_username and end_proxy.auth_password
                    else {}
                ),
            }
            print("End proxy request payload", proxy_payload)

            # Make sure the end proxy has configured correctly
            context = browser.new_context(
                proxy=proxy_payload,
            )
            page = context.new_page()
            page.goto("https://freeman.vc", timeout=5000)
            assert BeautifulSoup(page.content()).text.strip() == "Test content"

            # Make sure the middle proxy routes through the end proxy correctly
            context = browser.new_context(
                proxy={
                    "server": middle_proxy.base_url_proxy,
                }
            )
            page = context.new_page()
            page.goto("https://freeman.vc", timeout=5000)
            assert BeautifulSoup(page.content()).text.strip() == "Test content"
